function collisionstate = detectcollision()
% DETECTCOLLISION is an internal rountine of mat3d for collision detection.
%  The data strucures used in detectcollision is setup in the main script
%  mat3d.  Depending on the state of the collision detection feature, M3TRANS
%  or M3ROTATE calls detectcollision as appropriate. Use DETECTCONTACT to 
%  turn on/off collision detection feature.
%
%  DETECTCOLLISION returns a '1' if the objects being tested are in contact
%  and returns '0' if the objects are not in contact.
%
%  See also MAT3D, DETECTCONTACT, M3ROTATE, M3TRANS, SEL

% Copyright (c)1999, Nanping R. Lo nanping@ee.ucla.edu

global istruct;
global numstruct;
global m3debug;
global selectedObj;
global detectcontact;
global createArrowHandle;
global arrowhdl;
global comparison_cnt;

if (nargin ~= 0)
   disp('M3D: detectcollision requires no input.');
   return;
end

collisionstate = 0;

if (detectcontact == 0)
   return;   
end

if (createArrowHandle)
   arrowhdl = arrow([0 0 0],[0 0 0], 30, 80,[], 4,'CrossDir',[0 1 0],...
      'EdgeColor','r','FaceColor','none','LineWidth',1,'visible','off');
   createArrowHandle = 0;   
end

contact = 0;
comparison_cnt = 0;

%start_collision_detection = clock

if istruct(selectedObj).linkstart == 0
   contact = check_contact_for_structure(selectedObj);
else   
   linknumber = istruct(selectedObj).linkstart;
   for i=1:numstruct
      arraysize = size(istruct(i).link);
      if linknumber <= arraysize(2)
         if istruct(i).link(linknumber)
            contact = check_contact_for_structure(i);
            break
         end
      end
   end   
end

if contact == 0
   set(arrowhdl, 'visible', 'off');
end

collisionstate = contact;
%end_collision_detection = clock

return;

%------------------------------------------------------------------------%

function contact = check_contact_for_structure(i)

global numstruct;
global istruct;
global m3debug;

contact = 0;

ixpts = get(istruct(i).bbox, 'xdata');
iypts = get(istruct(i).bbox, 'ydata');
izpts = get(istruct(i).bbox, 'zdata');

[ixmax ixmin iymax iymin izmax izmin] = getmaxmin(ixpts, iypts, izpts);

% 1st check collision on the bounding box level
for k = 1:numstruct
   
   bbintersect = 0;
   
   %dont compare against self.
   if (k ~= i)
      kxpts = get(istruct(k).bbox, 'xdata');
      kypts = get(istruct(k).bbox, 'ydata');
      kzpts = get(istruct(k).bbox, 'zdata');
      
      [kxmax kxmin kymax kymin kzmax kzmin] = getmaxmin(kxpts, kypts, kzpts);
      
      maxminintersect = maxmintest(ixmax, ixmin, iymax, iymin, izmax, izmin, ...
                                   kxmax, kxmin, kymax, kymin, kzmax, kzmin);
                                
      if maxminintersect
         %disp ([k i]);                          
         for edge_number = 1:12
         
            [p1, p2] = set_segment_endpts(edge_number, kxpts, kypts, kzpts);
         
            for faces = 1:6
               idx  = 1 + 4*(faces-1);
               pa = [ixpts(idx)   iypts(idx)   izpts(idx)  ];
               pb = [ixpts(idx+1) iypts(idx+1) izpts(idx+1)];
               pc = [ixpts(idx+2) iypts(idx+2) izpts(idx+2)];
               pd = [ixpts(idx+3) iypts(idx+3) izpts(idx+3)];

              [bbintersect, intcoord] = segmentfacet(pa, pb, pc, pd, p1, p2);
      
               if (bbintersect)
                  contact = check_collision_between_structs(i,k);
                  if contact
                     return
                  end
                  break
               end
            end
         
            if (bbintersect)
               break
            end
         end 
      end
      
   end
end

return

%------------------------------------------------------------------------%

function contact = check_collision_between_structs(i,k)

global numstruct;
global istruct;
global ibox;
global m3debug;
global coordmax;
global arrowhdl;
global comparison_cnt;

contact = 0;

for j=istruct(i).firstbox:istruct(i).lastbox

   ixpts = get(ibox(j).hdl, 'xdata');
   iypts = get(ibox(j).hdl, 'ydata');
   izpts = get(ibox(j).hdl, 'zdata');
   
   [ixmax ixmin iymax iymin izmax izmin] = getmaxmin(ixpts, iypts, izpts);

   for m=istruct(k).firstbox:istruct(k).lastbox
      kxpts = get(ibox(m).hdl, 'xdata');
      kypts = get(ibox(m).hdl, 'ydata');
      kzpts = get(ibox(m).hdl, 'zdata');
      
      [kxmax kxmin kymax kymin kzmax kzmin] = getmaxmin(kxpts, kypts, kzpts);
      
      maxminintersect = maxmintest(ixmax, ixmin, iymax, iymin, izmax, izmin, ...
                                   kxmax, kxmin, kymax, kymin, kzmax, kzmin);
                                
      if maxminintersect
         %disp ([k i]);                          
         for edge_number = 1:12   %only do 6 out 12 edges
            
            [p1, p2] = set_segment_endpts(edge_number, kxpts, kypts, kzpts);               
            
            % since the height of the box is insignificant, could only do top & bottom   
            % faces of a box (omiting the 4 side faces).
            for faces = 1:6
               idx  = 1 + 4*(faces-1);
               pa = [ixpts(idx)   iypts(idx)   izpts(idx)  ];
               pb = [ixpts(idx+1) iypts(idx+1) izpts(idx+1)];
               pc = [ixpts(idx+2) iypts(idx+2) izpts(idx+2)];
               pd = [ixpts(idx+3) iypts(idx+3) izpts(idx+3)];            
               
               comparison_cnt = comparison_cnt +1;
               %mystats = [i k j m edge_number faces comparison_cnt]
               
               [intersect, intpt] = segmentfacet(pa, pb, pc, pd, p1, p2);
               
               if (intersect)
                  tmpstr = sprintf('between %d & %d at (%.1f %.1f %.1f)', ...
                                    k, i, intpt(1), intpt(2), intpt(3));
                  disp (['M3D: collision detected ' tmpstr]);
                  arrowstart = [intpt(1)+coordmax/15 intpt(2) intpt(3)];
                  arrow(arrowhdl, 'start', arrowstart, 'stop',intpt, 'visible','on');
                  contact = 1;
                  return
               end
               
            end %faces loop
         end %edge_number loop
      end %if maxmintest
   end
end

return

%------------------------------------------------------------------------%

function  [p1, p2] = set_segment_endpts(edge_number, xedge, yedge, zedge)

   switch edge_number
      case 1
         p1 = [xedge(1) yedge(1) zedge(1)];
         p2 = [xedge(2) yedge(2) zedge(2)];
      case 2
         p1 = [xedge(1) yedge(1) zedge(1)];
         p2 = [xedge(4) yedge(4) zedge(4)];  
      case 3
         p1 = [xedge(1) yedge(1) zedge(1)];
         p2 = [xedge(10)yedge(10)zedge(10)];
      case 4
         p1 = [xedge(7) yedge(7) zedge(7)];
         p2 = [xedge(3) yedge(3) zedge(3)];
      case 5
         p1 = [xedge(7) yedge(7) zedge(7)];
         p2 = [xedge(6) yedge(6) zedge(6)];
      case 6
         p1 = [xedge(7) yedge(7) zedge(7)];
         p2 = [xedge(11)yedge(11)zedge(11)];
      case 7
         p1 = [xedge(2) yedge(2) zedge(2)];
         p2 = [xedge(6) yedge(6) zedge(6)];
      case 8
         p1 = [xedge(2) yedge(2) zedge(2)];
         p2 = [xedge(3) yedge(3) zedge(3)];
      case 9
         p1 = [xedge(3) yedge(3) zedge(3)];
         p2 = [xedge(4) yedge(4) zedge(4)];
      case 10
         p1 = [xedge(6) yedge(6) zedge(6)];
         p2 = [xedge(10)yedge(10)zedge(10)];
      case 11
         p1 = [xedge(4) yedge(4) zedge(4)];
         p2 = [xedge(11)yedge(11)zedge(11)];
      case 12
         p1 = [xedge(10)yedge(10)zedge(10)];
         p2 = [xedge(11)yedge(11)zedge(11)];
      otherwise
         p1 = [xedge(1) yedge(1) zedge(1)];
         p2 = [xedge(7) yedge(7) zedge(7)];      
   end
      
return 

%------------------------------------------------------------------------%

         
function  [xmax, xmin, ymax, ymin, zmax, zmin] = getmaxmin(xpts, ypts, zpts)

xmax = xpts(1);
xmin = xpts(1);
ymax = ypts(1);
ymin = ypts(1);
zmax = zpts(1);
zmin = zpts(1);

for j = 1:24
   xmax = max (xmax, xpts(j));
   ymax = max (ymax, ypts(j));
   zmax = max (zmax, zpts(j));
   xmin = min (xmin, xpts(j));
   ymin = min (ymin, ypts(j));
   zmin = min (zmin, zpts(j));   
end

return

%------------------------------------------------------------------------%

function intersect = maxmintest(ixmax, ixmin, iymax, iymin, izmax, izmin, kxmax, kxmin, kymax, kymin, kzmax, kzmin)

intersect = 0;
xinter    = 0;
yinter    = 0;
zinter    = 0;

if (ixmax >= kxmax & ixmin <= kxmax)  | (ixmax >= kxmin & ixmin <= kxmin) 
   xinter = 1;
elseif (kxmax >= ixmax & kxmin <= ixmax)  | (kxmax >= ixmin & kxmin <= ixmin) 
   xinter = 1;
end

if (iymax >= kymax & iymin <= kymax)  | (iymax >= kymin & iymin <= kymin) 
   yinter = 1;
elseif (kymax >= iymax & kymin <= iymax)  | (kymax >= iymin & kymin <= iymin) 
   yinter = 1;
end

if (izmax >= kzmax & izmin <= kzmax)  | (izmax >= kzmin & izmin <= kzmin) 
   zinter = 1;
elseif (kzmax >= izmax & kzmin <= izmax)  | (kzmax >= izmin & kzmin <= izmin) 
   zinter = 1;
end

if xinter & yinter & zinter
   intersect = 1;
end

return