function [hit, coord, reflectdir] = rayfacet (pa, pb, pc, pd, origin, dir)
% RAYFACET is an internal rountine of mat3d.
%
%  determine whether or not the line segment p1,p2 intersects the 3 
%  vertex facet bounded by pa,pb,pc,pd
%  return true/false and the intersection point p
%
%   The equation of the line is p = p1 + mu (p2 - p1)
%
%  See also MAT3D, RAYSOURCE, RAYTRACING

% Copyright (c)1999, Nanping R. Lo nanping@ee.ucla.edu
%

if nargin ~= 6, error('rayfacet accepts 6 input.'); end
hit = [];
coord = origin;
reflectdir = dir;

% Calculate the parameters for the plane  (only use 3 points to determine the plane
% The equation of the plane is a x + b y + c z + d = 0
%                      ==>   n.x x + n.y y + n.z z + d = 0
% or (a,b,c) = (pb - pa)cross(pc - pa)

%vertices = [pa; pb; pc; pd];

nvec(1) = (pb(2) - pa(2))*(pc(3) - pa(3)) - (pb(3) - pa(3))*(pc(2) - pa(2));
nvec(2) = (pb(3) - pa(3))*(pc(1) - pa(1)) - (pb(1) - pa(1))*(pc(3) - pa(3));
nvec(3) = (pb(1) - pa(1))*(pc(2) - pa(2)) - (pb(2) - pa(2))*(pc(1) - pa(1));

epsilon = 0.001;

if (abs(norm(nvec)) < epsilon)
   n = [0 0 0];
else
   n = nvec/norm (nvec);
end

d = -n(1)*pa(1) - n(2)*pa(2) - n(3)*pa(3);

% Calculate the position on the line that intersects the plane
denom = n(1)*dir(1) + n(2)*dir(2) + n(3)*dir(3);

if (abs(denom) < epsilon)  
   % line and plane are parallel -> don't intersect
   hit = 0;
   return;
end

mu = - (d + n(1)*origin(1) + n(2)*origin(2) + n(3)*origin(3)) / denom;
coord(1) = origin(1) + mu * dir(1);
coord(2) = origin(2) + mu * dir(2);
coord(3) = origin(3) + mu * dir(3);

if (mu < 0)   % intersection not along ray 
   hit = 0;
   return;
end

% now determine whether or not the intersection point is bounded by pa,pb,pc
pa1vec(1) = pa(1) - coord(1);
pa1vec(2) = pa(2) - coord(2);
pa1vec(3) = pa(3) - coord(3);
pa1 = pa1vec/norm(pa1vec);

pa2vec(1) = pb(1) - coord(1);
pa2vec(2) = pb(2) - coord(2);
pa2vec(3) = pb(3) - coord(3);
pa2 = pa2vec/norm(pa2vec);

pa3vec(1) = pc(1) - coord(1);
pa3vec(2) = pc(2) - coord(2);
pa3vec(3) = pc(3) - coord(3);
pa3 = pa3vec/norm(pa3vec);

pa4vec(1) = pd(1) - coord(1);
pa4vec(2) = pd(2) - coord(2);
pa4vec(3) = pd(3) - coord(3);
pa4 = pa4vec/norm(pa4vec);

%compute the interior angles of the triangle fromed by pa,pb,pc 
a1 = pa1(1)*pa2(1) + pa1(2)*pa2(2) + pa1(3)*pa2(3);
a2 = pa2(1)*pa3(1) + pa2(2)*pa3(2) + pa2(3)*pa3(3);
a3 = pa3(1)*pa1(1) + pa3(2)*pa1(2) + pa3(3)*pa1(3);

%compute the interior angles of the triangle fromed by pa,pc,pd 
a4 = pa4(1)*pa1(1) + pa4(2)*pa1(2) + pa4(3)*pa1(3);
a5 = pa4(1)*pa3(1) + pa4(2)*pa3(2) + pa4(3)*pa3(3);
a6 = pa3(1)*pa1(1) + pa3(2)*pa1(2) + pa3(3)*pa1(3);

total_1 = (acos(a1) + acos(a2) + acos(a3)) * 180/pi;
total_2 = (acos(a4) + acos(a5) + acos(a6)) * 180/pi;

if (abs(total_1 - 360) > epsilon) & (abs(total_2 - 360) > epsilon)
  hit = 0;
  return;
end

reflectdir = dir - 2* dot(dir,n) * n ;
hit = 1;

return;

